@@ -1,6 +1,5 @@
 from pkg_resources import get_distribution, DistributionNotFound
 import os.path
-from .stomp import stomp  # noqa: F401
 from .stump import stump, _stump, _get_first_stump_profile, _get_QT  # noqa: F401
 from .stumped import stumped  # noqa: F401
 from .mstump import (  # noqa: F401
@@ -9,11 +8,13 @@ from .mstump import (  # noqa: F401
     _get_first_mstump_profile,
     _get_multi_QT,
     _multi_mass,
+    _apply_include,
 )
 from .mstumped import mstumped  # noqa: F401
 from .chains import atsc, allc  # noqa: F401
 from .floss import floss, fluss, _nnmark, _iac, _cac, _rea  # noqa: F401
-from .scrimp import scrimp  # noqa: F401
+from .scrump import scrump  # noqa: F401
+from .stumpi import stumpi  # noqa: F401
 from numba import cuda
 
 if cuda.is_available():
