@@ -8,6 +8,8 @@ import scipy.signal
 import tempfile
 import math
 
+from . import config
+
 try:
     from numba.cuda.cudadrv.driver import _raise_driver_not_found
 except ImportError:
@@ -16,23 +18,21 @@ except ImportError:
 
 def driver_not_found(*args, **kwargs):  # pragma: no cover
     """
-    Helper function to raise CudaSupportError driver not found error
+    Helper function to raise CudaSupportError driver not found error.
     """
-
     _raise_driver_not_found()
 
 
 def get_pkg_name():  # pragma: no cover
     """
-    Return package name
+    Return package name.
     """
-
     return __name__.split(".")[0]
 
 
 def rolling_window(a, window):
     """
-    Use strides to generate rolling/sliding windows for a numpy array
+    Use strides to generate rolling/sliding windows for a numpy array.
 
     Parameters
     ----------
@@ -47,7 +47,7 @@ def rolling_window(a, window):
     output : ndarray
         This will be a new view of the original input array.
     """
-
+    a = np.asarray(a)
     shape = a.shape[:-1] + (a.shape[-1] - window + 1, window)
     strides = a.strides + (a.strides[-1],)
 
@@ -80,16 +80,15 @@ def z_norm(a, axis=0):
 
 def check_nan(a):  # pragma: no cover
     """
-    Check if the array contains NaNs
+    Check if the array contains NaNs.
 
     Raises
     ------
     ValueError
         If the array contains a NaN
     """
-
     if np.any(np.isnan(a)):
-        msg = f"Input array contains one or more NaNs"
+        msg = "Input array contains one or more NaNs"
         raise ValueError(msg)
 
     return
@@ -104,7 +103,6 @@ def check_dtype(a, dtype=np.floating):  # pragma: no cover
     TypeError
         If the array type does not match `dtype`
     """
-
     if not np.issubdtype(a.dtype, dtype):
         msg = f"{dtype} type expected but found {a.dtype}"
         raise TypeError(msg)
@@ -131,7 +129,6 @@ def transpose_dataframe(a):  # pragma: no cover
     output : a
         If a is a Pandas `DataFrame` then return `a.T`. Otherwise, return `a`
     """
-
     if type(a).__name__ == "DataFrame":
         return a.T
 
@@ -156,7 +153,6 @@ def are_arrays_equal(a, b):  # pragma: no cover
     output : bool
         Returns `True` if the arrays are equal and `False` otherwise.
     """
-
     if id(a) == id(b):
         return True
 
@@ -187,7 +183,6 @@ def are_distances_too_small(a, threshold=10e-6):  # pragma: no cover
         Returns `True` if the matrix profile distances are all below the
         threshold and `False` if they are all above the threshold.
     """
-
     if a.mean() < threshold or np.all(a < threshold):
         return True
 
@@ -196,6 +191,7 @@ def are_distances_too_small(a, threshold=10e-6):  # pragma: no cover
 
 def check_window_size(m):
     """
+    Check the window size and ensure that it is greater than or equal to 3
 
     Parameters
     ----------
@@ -252,7 +248,6 @@ def sliding_dot_product(Q, T):
 
     Padding is done automatically in fftconvolve step
     """
-
     n = T.shape[0]
     m = Q.shape[0]
     Qr = np.flipud(Q)  # Reverse/flip Q
@@ -261,10 +256,10 @@ def sliding_dot_product(Q, T):
     return QT.real[m - 1 : n]
 
 
-def compute_mean_std(T, m, num_chunks=1, max_iter=10):
+def compute_mean_std(T, m):
     """
     Compute the sliding mean and standard deviation for the array `T` with
-    a window size of `m` by splitting up T in `n_chunks`
+    a window size of `m`
 
     Parameters
     ----------
@@ -274,13 +269,6 @@ def compute_mean_std(T, m, num_chunks=1, max_iter=10):
     m : int
         Window size
 
-    num_chunks : int
-        Number of chunks to use for the first iteration. If a Memory Error is raised,
-        this number is doubled and the computation tried again.
-
-    max_iter : int
-        Maximum number of iterations
-
     Returns
     -------
     M_T : ndarray
@@ -309,22 +297,29 @@ def compute_mean_std(T, m, num_chunks=1, max_iter=10):
     Note that Mueen's algorithm has an off-by-one bug where the
     sum for the first subsequence is omitted and we fixed that!
     """
+    num_chunks = config.STUMPY_MEAN_STD_NUM_CHUNKS
+    max_iter = config.STUMPY_MEAN_STD_MAX_ITER
+
     if T.ndim > 2:  # pragma nocover
         raise ValueError("T has to be one or two dimensional!")
 
     for iteration in range(max_iter):
         try:
             chunk_size = math.ceil((T.shape[-1] + 1) / num_chunks)
+            if chunk_size < m:
+                chunk_size = m
 
             mean_chunks = []
             std_chunks = []
             for chunk in range(num_chunks):
                 start = chunk * chunk_size
                 stop = min(start + chunk_size + m - 1, T.shape[-1])
+                if stop - start < m:
+                    break
 
-                tmp_mean = np.mean(rolling_window(T[start:stop], m), axis=T.ndim)
+                tmp_mean = np.mean(rolling_window(T[..., start:stop], m), axis=T.ndim)
                 mean_chunks.append(tmp_mean)
-                tmp_std = np.nanstd(rolling_window(T[start:stop], m), axis=T.ndim)
+                tmp_std = np.nanstd(rolling_window(T[..., start:stop], m), axis=T.ndim)
                 std_chunks.append(tmp_std)
 
             M_T = np.hstack(mean_chunks)
@@ -346,6 +341,62 @@ def compute_mean_std(T, m, num_chunks=1, max_iter=10):
         )
 
 
+@njit(fastmath=True)
+def _calculate_squared_distance(m, QT, μ_Q, σ_Q, M_T, Σ_T):
+    """
+    Compute a single squared distance given all scalar inputs. This function serves as
+    the single source of truth for how all distances should be calculated.
+
+    Parameters
+    ----------
+    m : int
+        Window size
+
+    QT : float
+        Dot product between `Q[i]` and `T[i]`
+
+    μ_Q : float
+        Mean of `Q[i]`
+
+    σ_Q : float
+        Standard deviation of `Q[i]`
+
+    M_T : float
+        Sliding mean of `T[i]`
+
+    Σ_T : float
+        Sliding standard deviation of `T[i]`
+
+    Returns
+    -------
+    D_squared : float
+        Squared distance
+
+    Notes
+    -----
+    `DOI: 10.1109/ICDM.2016.0179 \
+    <https://www.cs.ucr.edu/~eamonn/PID4481997_extend_Matrix%20Profile_I.pdf>`__
+
+    See Equation on Page 4
+    """
+    threshold = 1e-10
+    if np.isinf(M_T) or np.isinf(μ_Q):
+        D_squared = np.inf
+    else:
+        if σ_Q < threshold or Σ_T < threshold:
+            D_squared = m
+        else:
+            denom = m * σ_Q * Σ_T
+            if np.abs(denom) < threshold:  # pragma nocover
+                denom = threshold
+            D_squared = np.abs(2 * m * (1.0 - (QT - m * μ_Q * M_T) / denom))
+
+        if σ_Q < threshold and Σ_T < threshold:
+            D_squared = 0
+
+    return D_squared
+
+
 @njit(parallel=True, fastmath=True)
 def _calculate_squared_distance_profile(m, QT, μ_Q, σ_Q, M_T, Σ_T):
     """
@@ -373,8 +424,8 @@ def _calculate_squared_distance_profile(m, QT, μ_Q, σ_Q, M_T, Σ_T):
 
     Returns
     -------
-    output : ndarray
-        Distance profile squared
+    D_squared : ndarray
+        Squared distance profile
 
     Notes
     -----
@@ -383,27 +434,11 @@ def _calculate_squared_distance_profile(m, QT, μ_Q, σ_Q, M_T, Σ_T):
 
     See Equation on Page 4
     """
-
     k = M_T.shape[0]
     D_squared = np.empty(k)
 
-    threshold = 1e-10
     for i in prange(k):
-        if np.isinf(M_T[i]) or np.isinf(μ_Q):
-            D_squared[i] = np.inf
-        else:
-            if σ_Q < threshold or Σ_T[i] < threshold:
-                D_squared[i] = m
-            else:
-                denom = m * σ_Q * Σ_T[i]
-                if np.abs(denom) < threshold:  # pragma nocover
-                    denom = threshold
-                D_squared[i] = np.abs(
-                    2 * m * (1.0 - (QT[i] - m * μ_Q * M_T[i]) / denom)
-                )
-
-            if σ_Q < threshold and Σ_T[i] < threshold:
-                D_squared[i] = 0
+        D_squared[i] = _calculate_squared_distance(m, QT[i], μ_Q, σ_Q, M_T[i], Σ_T[i])
 
     return D_squared
 
@@ -445,7 +480,6 @@ def calculate_distance_profile(m, QT, μ_Q, σ_Q, M_T, Σ_T):
 
     See Equation on Page 4
     """
-
     D_squared = _calculate_squared_distance_profile(m, QT, μ_Q, σ_Q, M_T, Σ_T)
 
     return np.sqrt(D_squared)
@@ -519,9 +553,56 @@ def mueen_calculate_distance_profile(Q, T):
     return np.sqrt(D)
 
 
+@njit(fastmath=True)
+def _mass(Q, T, QT, μ_Q, σ_Q, M_T, Σ_T):
+    """
+    A Numba JIT compiled algorithm for computing the distance profile using the MASS
+    algorithm.
+
+    Parameters
+    ----------
+    Q : ndarray
+        Query array or subsequence
+
+    T : ndarray
+        Time series or sequence
+
+    QT : ndarray
+        The sliding dot product of Q and T
+
+    μ_Q : float
+        The scalar mean of Q
+
+    σ_Q : float
+        The scalar standard deviation of Q
+
+    M_T : ndarray
+        Sliding mean of `T`
+
+    Σ_T : ndarray
+        Sliding standard deviation of `T`
+
+    Notes
+    -----
+    `DOI: 10.1109/ICDM.2016.0179 \
+    <https://www.cs.ucr.edu/~eamonn/PID4481997_extend_Matrix%20Profile_I.pdf>`__
+
+    See Table II
+
+    Note that Q, T are not directly required to calculate D
+
+    Note: Unlike the Matrix Profile I paper, here, M_T, Σ_T can be calculated
+    once for all subsequences of T and passed in so the redundancy is removed
+    """
+    m = Q.shape[0]
+
+    return calculate_distance_profile(m, QT, μ_Q, σ_Q, M_T, Σ_T)
+
+
 def mass(Q, T, M_T=None, Σ_T=None):
     """
-    Compute the distance profile using the MASS algorithm
+    Compute the distance profile using the MASS algorithm. This is a convenience
+    wrapper around the Numba JIT compiled `_mass` function.
 
     Parameters
     ----------
@@ -554,36 +635,97 @@ def mass(Q, T, M_T=None, Σ_T=None):
     Note: Unlike the Matrix Profile I paper, here, M_T, Σ_T can be calculated
     once for all subsequences of T and passed in so the redundancy is removed
     """
-
     Q = np.asarray(Q)
     check_dtype(Q)
+    m = Q.shape[0]
 
     if Q.ndim != 1:  # pragma: no cover
         raise ValueError(f"Q is {Q.ndim}-dimensional and must be 1-dimensional. ")
-    m = Q.shape[0]
 
     T = np.asarray(T)
     check_dtype(T)
+    n = T.shape[0]
 
     if T.ndim != 1:  # pragma: no cover
         raise ValueError(f"T is {T.ndim}-dimensional and must be 1-dimensional. ")
-    n = T.shape[0]
 
     distance_profile = np.empty(n - m + 1)
     if np.any(np.isnan(Q)):
         distance_profile[:] = np.inf
     else:
+        if M_T is None or Σ_T is None:
+            T, M_T, Σ_T = preprocess(T, m)
+
         QT = sliding_dot_product(Q, T)
         μ_Q, σ_Q = compute_mean_std(Q, m)
-        if M_T is None or Σ_T is None:
-            M_T, Σ_T = compute_mean_std(T, m)
-        distance_profile = calculate_distance_profile(
-            m, QT, μ_Q.item(0), σ_Q.item(0), M_T, Σ_T
-        )
+        μ_Q = μ_Q[0]
+        σ_Q = σ_Q[0]
+        distance_profile[:] = _mass(Q, T, QT, μ_Q, σ_Q, M_T, Σ_T)
 
     return distance_profile
 
 
+@njit(fastmath=True)
+def apply_exclusion_zone(D, idx, excl_zone):
+    """
+    Apply an exclusion zone to an array (inplace), i.e. set all values
+    to np.inf in a window around a given index.
+
+    All values in D in [idx - excl_zone, idx + excl_zone] (endpoints included)
+    will be set to np.inf.
+
+    Parameters
+    ----------
+    D : ndarray
+        The array you want to apply the exclusion zone to
+
+    idx : int
+        The index around which the window should be centered
+
+    excl_zone : int
+        Size of the exclusion zone.
+    """
+    zone_start = max(0, idx - excl_zone)
+    zone_stop = min(D.shape[-1], idx + excl_zone)
+    D[..., zone_start : zone_stop + 1] = np.inf
+
+
+def preprocess(T, m):
+    """
+    Creates a copy of the time series where all NaN and inf values
+    are replaced with zero. Also computes mean and standard deviation
+    for every subsequence. Every subsequence that contains at least
+    one NaN or inf value, will have a mean of np.inf. For the standard
+    deviation these values are ignored. If all values are illegal, the
+    standard deviation will be 0 (see `core.compute_mean_std`)
+
+    Parameters
+    ----------
+    T : ndarray
+        Time series or sequence
+
+    m : int
+        Window size
+
+    Returns
+    -------
+    T : ndarray
+        Modified time series
+    M_T : ndarray
+        Rolling mean
+    Σ_T : ndarray
+        Rolling standard deviation
+    """
+    T = T.copy()
+    T = np.asarray(T)
+
+    T[np.isinf(T)] = np.nan
+    M_T, Σ_T = compute_mean_std(T, m)
+    T[np.isnan(T)] = 0
+
+    return T, M_T, Σ_T
+
+
 def array_to_temp_file(a):
     """
     Write an ndarray to a file
@@ -598,7 +740,6 @@ def array_to_temp_file(a):
     fname : str
         The output file name
     """
-
     fname = tempfile.NamedTemporaryFile(delete=False)
     fname = fname.name + ".npy"
     np.save(fname, a, allow_pickle=False)
@@ -606,4 +747,58 @@ def array_to_temp_file(a):
     return fname
 
 
+def _get_array_ranges(
+    a, n_chunks, custom_cumsum=None, custom_insert=None, truncate=False
+):
+    """
+    Given an input array, split it into `n_chunks`.
+
+    Parameters
+    ----------
+    a : ndarray
+        An array to be split
+
+    n_chunks : int
+        Number of chunks to split the array into
+
+    custom_cumsum : ndarray
+        Custom cumsum array
+
+    custom_insert : ndarray
+        Custom insertion array to be inserted into the cumsum array
+
+    truncate : bool
+        If `truncate=True`, truncate the rows of `array_ranges` if there are not enough
+        elements in `a` to be chunked up into `n_chunks`.  Otherwise, if
+        `truncate=False`, all extra chunks will have their start and stop indices set
+        to `a.shape[0]`.
+
+    Returns
+    -------
+    array_ranges : ndarray
+        A two column array where each row consists of a start and (exclusive) stop index
+        pair. The first column contains the start indices and the second column
+        contains the stop indices.
+    """
+    array_ranges = np.zeros((n_chunks, 2), np.int64)
+    if custom_cumsum is None:
+        custom_cumsum = a.cumsum() / a.sum()
+    if custom_insert is None:
+        custom_insert = np.linspace(0, 1, n_chunks, endpoint=False)[1:]
+    idx = 1 + np.searchsorted(custom_cumsum, custom_insert)
+    array_ranges[1:, 0] = idx  # Fill the first column with start indices
+    array_ranges[:-1, 1] = idx  # Fill the second column with exclusive stop indices
+    array_ranges[-1, 1] = a.shape[0]  # Handle the stop index for the final chunk
+
+    diff_idx = np.diff(idx)
+    if np.any(diff_idx == 0):
+        row_truncation_idx = np.argmin(diff_idx) + 2
+        array_ranges[row_truncation_idx:, 0] = a.shape[0]
+        array_ranges[row_truncation_idx - 1 :, 1] = a.shape[0]
+        if truncate:
+            array_ranges = array_ranges[:row_truncation_idx]
+
+    return array_ranges
+
+
 convolution = scipy.signal.fftconvolve  # Swap for other convolution function
