@@ -9,16 +9,14 @@ import os
 import numpy as np
 from numba import cuda
 
-from . import _get_QT
-from . import core
+from . import core, _get_QT, config
 
 logger = logging.getLogger(__name__)
-THREADS_PER_BLOCK = 512
 
 
 @cuda.jit(
     "(i8, f8[:], f8[:], i8,  f8[:], f8[:], f8[:], f8[:], f8[:],"
-    "f8[:], f8[:], i8, b1, i8, f8[:, :], f8[:, :], b1)"
+    "f8[:], f8[:], i8, b1, i8, f8[:, :], i8[:, :], b1)"
 )
 def _compute_and_update_PI_kernel(
     i,
@@ -115,10 +113,9 @@ def _compute_and_update_PI_kernel(
 
     See Table II, Figure 5, and Figure 6
     """
-
     start = cuda.grid(1)
     stride = cuda.gridsize(1)
-    threshold = 1e-7
+    threshold = 1e-10
 
     if i % 2 == 0:
         QT_out = QT_even
@@ -130,9 +127,6 @@ def _compute_and_update_PI_kernel(
     for j in range(start, QT_out.shape[0], stride):
         zone_start = max(0, j - excl_zone)
         zone_stop = min(k, j + excl_zone)
-        denom = m * σ_Q[i] * Σ_T[j]
-        if denom == 0:  # pragma: no cover
-            denom = 1e-10
 
         if compute_QT:
             QT_out[j] = (
@@ -140,11 +134,19 @@ def _compute_and_update_PI_kernel(
             )
 
             QT_out[0] = QT_first[i]
-        D = abs(2 * m * (1.0 - (QT_out[j] - m * μ_Q[i] * M_T[j]) / denom))
-        if σ_Q[i] < threshold or Σ_T[j] < threshold:  # pragma: no cover
-            D = m
-        if Σ_T[j] < threshold and σ_Q[i] < threshold:  # pragma: no cover
-            D = 0
+        if math.isinf(M_T[j]) or math.isinf(μ_Q[i]):
+            D = np.inf
+        else:
+            if σ_Q[i] < threshold or Σ_T[j] < threshold:
+                D = m
+            else:
+                denom = m * σ_Q[i] * Σ_T[j]
+                if math.fabs(denom) < threshold:  # pragma nocover
+                    denom = threshold
+                D = abs(2 * m * (1.0 - (QT_out[j] - m * μ_Q[i] * M_T[j]) / denom))
+
+            if σ_Q[i] < threshold and Σ_T[j] < threshold:
+                D = 0
 
         if ignore_trivial:
             if i <= zone_stop and i >= zone_start:
@@ -176,7 +178,6 @@ def _gpu_stump(
     k,
     ignore_trivial=True,
     range_start=1,
-    threads_per_block=THREADS_PER_BLOCK,
     device_id=0,
 ):
     """
@@ -239,10 +240,6 @@ def _gpu_stump(
         The starting index value along T_B for which to start the matrix
         profile calculation. Default is 1.
 
-    threads_per_block : int
-        The number of GPU threads to use for all kernels. The default value is
-        set in `THREADS_PER_BLOCK=512`.
-
     device_id : int
         The (GPU) device number to use. The default value is `0`.
 
@@ -285,7 +282,7 @@ def _gpu_stump(
 
     Note that left and right matrix profiles are only available for self-joins.
     """
-
+    threads_per_block = config.STUMPY_THREADS_PER_BLOCK
     blocks_per_grid = math.ceil(k / threads_per_block)
 
     T_A = np.load(T_A_fname, allow_pickle=False)
@@ -314,13 +311,12 @@ def _gpu_stump(
             device_Σ_T = cuda.to_device(Σ_T)
 
         profile = np.empty((k, 3))  # float64
-        indices = np.empty((k, 3))  # int64
+        indices = np.empty((k, 3), dtype=np.int64)  # int64
 
         profile[:] = np.inf
         indices[:, :] = -1
         device_profile = cuda.to_device(profile)
         device_indices = cuda.to_device(indices)
-
         _compute_and_update_PI_kernel[blocks_per_grid, threads_per_block](
             range_start - 1,
             device_T_A,
@@ -372,14 +368,7 @@ def _gpu_stump(
     return profile_fname, indices_fname
 
 
-def gpu_stump(
-    T_A,
-    m,
-    T_B=None,
-    ignore_trivial=True,
-    threads_per_block=THREADS_PER_BLOCK,
-    device_id=0,
-):
+def gpu_stump(T_A, m, T_B=None, ignore_trivial=True, device_id=0):
     """
     Compute the matrix profile with GPU-STOMP
 
@@ -402,10 +391,6 @@ def gpu_stump(
         Set to `True` if this is a self-join. Otherwise, for AB-join, set this
         to `False`. Default is `True`.
 
-    threads_per_block : int
-        The number of GPU threads to use for all kernels. The default value is
-        set in `THREADS_PER_BLOCK=512`.
-
     device_id : int or list
         The (GPU) device number to use. The default value is `0`. A list of
         valid device ids (int) may also be provided for parallel GPU-STUMP
@@ -422,7 +407,6 @@ def gpu_stump(
 
     Notes
     -----
-
     `DOI: 10.1109/ICDM.2016.0085 \
     <https://www.cs.ucr.edu/~eamonn/STOMP_GPU_final_submission_camera_ready.pdf>`__
 
@@ -449,10 +433,18 @@ def gpu_stump(
 
     Note that left and right matrix profiles are only available for self-joins.
     """
+    if T_B is None:  # Self join!
+        T_B = T_A
+        ignore_trivial = True
 
-    T_A = np.asarray(T_A)
-    core.check_dtype(T_A)
-    core.check_nan(T_A)
+    # Swap T_A and T_B for GPU implementation
+    # This keeps the API identical to and compatible with `stumpy.stump`
+    tmp_T = T_A
+    T_A = T_B
+    T_B = tmp_T
+
+    T_A, M_T, Σ_T = core.preprocess(T_A, m)
+    T_B, μ_Q, σ_Q = core.preprocess(T_B, m)
 
     if T_A.ndim != 1:  # pragma: no cover
         raise ValueError(
@@ -460,19 +452,15 @@ def gpu_stump(
             "For multidimensional STUMP use `stumpy.mstump` or `stumpy.mstumped`"
         )
 
-    if T_B is None:  # Self join!
-        T_B = T_A
-        ignore_trivial = True
-    T_B = np.asarray(T_B)
-
-    core.check_dtype(T_B)
-    core.check_nan(T_B)
     if T_B.ndim != 1:  # pragma: no cover
         raise ValueError(
             f"T_B is {T_B.ndim}-dimensional and must be 1-dimensional. "
             "For multidimensional STUMP use `stumpy.mstump` or `stumpy.mstumped`"
         )
 
+    core.check_dtype(T_A)
+    core.check_dtype(T_B)
+
     core.check_window_size(m)
 
     if ignore_trivial is False and core.are_arrays_equal(T_A, T_B):  # pragma: no cover
@@ -483,19 +471,11 @@ def gpu_stump(
         logger.warning("Arrays T_A, T_B are not equal, which implies an AB-join.")
         logger.warning("Try setting `ignore_trivial = False`.")
 
-    # Swap T_A and T_B for GPU implementation
-    # This keeps the API identical to and compatible with `stumpy.stump`
-    tmp_T = T_A
-    T_A = T_B
-    T_B = tmp_T
     n = T_B.shape[0]
     k = T_A.shape[0] - m + 1
     l = n - m + 1
     excl_zone = int(np.ceil(m / 4))  # See Definition 3 and Figure 3
 
-    M_T, Σ_T = core.compute_mean_std(T_A, m)
-    μ_Q, σ_Q = core.compute_mean_std(T_B, m)
-
     T_A_fname = core.array_to_temp_file(T_A)
     T_B_fname = core.array_to_temp_file(T_B)
     M_T_fname = core.array_to_temp_file(M_T)
@@ -559,7 +539,6 @@ def gpu_stump(
                     k,
                     ignore_trivial,
                     start + 1,
-                    threads_per_block,
                     device_ids[idx],
                 ),
             )
@@ -581,7 +560,6 @@ def gpu_stump(
                 k,
                 ignore_trivial,
                 start + 1,
-                threads_per_block,
                 device_ids[idx],
             )
 
