@@ -12,7 +12,7 @@ from . import core
 logger = logging.getLogger(__name__)
 
 
-def mstumped(dask_client, T, m):
+def mstumped(dask_client, T, m, include=None, discords=False):
     """
     Compute the multi-dimensional matrix profile with parallelized and
     distributed mSTOMP
@@ -38,6 +38,20 @@ def mstumped(dask_client, T, m):
     m : int
         Window size
 
+    include : list, ndarray
+        A list of (zero-based) indices corresponding to the dimensions in `T` that
+        must be included in the constrained multidimensional motif search.
+        For more information, see Section IV D in:
+
+        `DOI: 10.1109/ICDM.2017.66 \
+        <https://www.cs.ucr.edu/~eamonn/Motif_Discovery_ICDM.pdf>`__
+
+    discords : bool
+        When set to `True`, this reverses the distance matrix which results in a
+        multi-dimensional matrix profile that favors larger matrix profile values
+        (i.e., discords) rather than smaller values (i.e., motifs). Note that indices
+        in `include` are still maintained and respected.
+
     Returns
     -------
     P : ndarray
@@ -51,35 +65,37 @@ def mstumped(dask_client, T, m):
 
     Notes
     -----
-
     `DOI: 10.1109/ICDM.2017.66 \
     <https://www.cs.ucr.edu/~eamonn/Motif_Discovery_ICDM.pdf>`__
 
     See mSTAMP Algorithm
     """
+    T_A = core.transpose_dataframe(T)
+    T_B = T_A
 
-    T_A = np.asarray(core.transpose_dataframe(T)).copy()
-    T_B = T_A.copy()
+    T_A, M_T, Σ_T = core.preprocess(T_A, m)
+    T_B, μ_Q, σ_Q = core.preprocess(T_B, m)
 
-    T_A[np.isinf(T_A)] = np.nan
-    T_B[np.isinf(T_B)] = np.nan
-
-    core.check_dtype(T_A)
     if T_A.ndim <= 1:  # pragma: no cover
         err = f"T is {T_A.ndim}-dimensional and must be at least 1-dimensional"
         raise ValueError(f"{err}")
 
+    core.check_dtype(T_A)
+    core.check_dtype(T_B)
+
     core.check_window_size(m)
 
-    d, n = T_A.shape
+    if include is not None:
+        include = np.asarray(include)
+        _, idx = np.unique(include, return_index=True)
+        if include.shape[0] != idx.shape[0]:  # pragma: no cover
+            logger.warning("Removed repeating indices in `include`")
+            include = include[np.sort(idx)]
+
+    d, n = T_B.shape
     k = n - m + 1
     excl_zone = int(np.ceil(m / 4))  # See Definition 3 and Figure 3
 
-    M_T, Σ_T = core.compute_mean_std(T_A, m)
-    μ_Q, σ_Q = core.compute_mean_std(T_B, m)
-
-    T_A[np.isnan(T_A)] = 0
-
     P = np.empty((d, k), dtype="float64")
     I = np.empty((d, k), dtype="int64")
 
@@ -90,11 +106,9 @@ def mstumped(dask_client, T, m):
 
     for i, start in enumerate(range(0, k, step)):
         P[:, start], I[:, start] = _get_first_mstump_profile(
-            start, T_A, T_B, m, excl_zone, M_T, Σ_T
+            start, T_A, T_B, m, excl_zone, M_T, Σ_T, μ_Q, σ_Q, include, discords
         )
 
-    T_B[np.isnan(T_B)] = 0
-
     # Scatter data to Dask cluster
     T_A_future = dask_client.scatter(T_A, broadcast=True)
     M_T_future = dask_client.scatter(M_T, broadcast=True)
@@ -133,6 +147,8 @@ def mstumped(dask_client, T, m):
                 σ_Q_future,
                 k,
                 start + 1,
+                include,
+                discords,
             )
         )
 
