@@ -12,12 +12,14 @@ from . import core, stamp
 logger = logging.getLogger(__name__)
 
 
-def _get_first_stump_profile(start, T_A, T_B, m, excl_zone, M_T, Σ_T, ignore_trivial):
+def _get_first_stump_profile(
+    start, T_A, T_B, m, excl_zone, M_T, Σ_T, μ_Q, σ_Q, ignore_trivial
+):
     """
     Compute the matrix profile, matrix profile index, left matrix profile
     index, and right matrix profile index for given window within the times
     series or sequence that is denote by the `start` index. Essentially, this
-    is a convenience wrapper around `stamp.mass`
+    is a convenience wrapper around `stamp._mass_PI`
 
     Parameters
     ----------
@@ -44,13 +46,22 @@ def _get_first_stump_profile(start, T_A, T_B, m, excl_zone, M_T, Σ_T, ignore_tr
     Σ_T : ndarray
         Sliding standard deviation for `T_A`
 
+    μ_Q : ndarray
+        Sliding mean for `T_B`
+
+    σ_Q : ndarray
+        Sliding standard deviation for `T_B`
+
     ignore_trivial : bool
         `True` if this is a self join and `False` otherwise (i.e., AB-join).
 
     Returns
     -------
-    P : float64
-        Matrix profile for the window with index equal to `start`
+    P : Tuple[float64, float64, float64]
+        Matrix profile, left matrix profile index, and right matrix profile for the
+        window with index equal to `start`. The left and right matrix profile are
+        automatically set to `np.inf` for self-joins (i.e., when `ignore_trivial` is
+        set to `True`.
 
     I : Tuple[int64, int64, int64]
         Matrix profile index, left matrix profile index, and right matrix profile
@@ -58,23 +69,36 @@ def _get_first_stump_profile(start, T_A, T_B, m, excl_zone, M_T, Σ_T, ignore_tr
         profile indices are automatically set to `-1` for self-joins (i.e., when
         `ignore_trivial` is set to `True`.
     """
-
     # Handle first subsequence, add exclusionary zone
-    if ignore_trivial:
-        P, I = stamp.mass(T_B[start : start + m], T_A, M_T, Σ_T, start, excl_zone)
-        PL, IL = stamp.mass(
-            T_B[start : start + m], T_A, M_T, Σ_T, start, excl_zone, left=True
-        )
-        PR, IR = stamp.mass(
-            T_B[start : start + m], T_A, M_T, Σ_T, start, excl_zone, right=True
-        )
-    else:
-        P, I = stamp.mass(T_B[start : start + m], T_A, M_T, Σ_T)
-        # No left and right matrix profile available
+
+    if np.isinf(μ_Q[start]):
+        P = np.inf
+        PL = np.inf
+        PR = np.inf
+
+        I = -1
         IL = -1
         IR = -1
+    else:
+        if ignore_trivial:
+            P, I = stamp._mass_PI(
+                T_B[start : start + m], T_A, M_T, Σ_T, start, excl_zone
+            )
+            PL, IL = stamp._mass_PI(
+                T_B[start : start + m], T_A, M_T, Σ_T, start, excl_zone, left=True
+            )
+            PR, IR = stamp._mass_PI(
+                T_B[start : start + m], T_A, M_T, Σ_T, start, excl_zone, right=True
+            )
+        else:
+            P, I = stamp._mass_PI(T_B[start : start + m], T_A, M_T, Σ_T)
+            # No left and right matrix profile available
+            PL = np.inf
+            PR = np.inf
+            IL = -1
+            IR = -1
 
-    return P, (I, IL, IR)
+    return (P, PL, PR), (I, IL, IR)
 
 
 def _get_QT(start, T_A, T_B, m):
@@ -106,7 +130,6 @@ def _get_QT(start, T_A, T_B, m):
     QT_first : ndarray
          QT for the first window
     """
-
     QT = core.sliding_dot_product(T_B[start : start + m], T_A)
     QT_first = core.sliding_dot_product(T_A[:m], T_B)
 
@@ -224,7 +247,6 @@ def _stump(
 
     Note that left and right matrix profiles are only available for self-joins.
     """
-
     QT_odd = QT.copy()
     QT_even = QT.copy()
     profile = np.empty((range_stop - range_start,))  # float64
@@ -265,9 +287,7 @@ def _stump(
             )
 
         if ignore_trivial:
-            zone_start = max(0, i - excl_zone)
-            zone_stop = min(k, i + excl_zone)
-            D[zone_start : zone_stop + 1] = np.inf
+            core.apply_exclusion_zone(D, i, excl_zone)
 
         I = np.argmin(D)
         P = np.sqrt(D[I])
@@ -280,7 +300,7 @@ def _stump(
         if ignore_trivial and i > 0:
             IL = np.argmin(D[:i])
             PL = D[IL]
-        if PL == np.inf or zone_start <= IL < zone_stop:
+        if PL == np.inf:
             IL = -1
 
         IR = -1
@@ -288,7 +308,7 @@ def _stump(
         if ignore_trivial and i + 1 < D.shape[0]:
             IR = i + 1 + np.argmin(D[i + 1 :])
             PR = D[IR]
-        if PR == np.inf or zone_start <= IR < zone_stop:
+        if PR == np.inf:
             IR = -1
 
         # Only a part of the profile/indices array are passed
@@ -331,7 +351,6 @@ def stump(T_A, m, T_B=None, ignore_trivial=True):
 
     Notes
     -----
-
     `DOI: 10.1109/ICDM.2016.0085 \
     <https://www.cs.ucr.edu/~eamonn/STOMP_GPU_final_submission_camera_ready.pdf>`__
 
@@ -358,31 +377,26 @@ def stump(T_A, m, T_B=None, ignore_trivial=True):
 
     Note that left and right matrix profiles are only available for self-joins.
     """
+    if T_B is None:
+        T_B = T_A
+        ignore_trivial = True
+
+    T_A, M_T, Σ_T = core.preprocess(T_A, m)
+    T_B, μ_Q, σ_Q = core.preprocess(T_B, m)
 
-    T_A = np.asarray(T_A)
     if T_A.ndim != 1:  # pragma: no cover
         raise ValueError(
             f"T_A is {T_A.ndim}-dimensional and must be 1-dimensional. "
             "For multidimensional STUMP use `stumpy.mstump` or `stumpy.mstumped`"
         )
-    n = T_A.shape[0]
 
-    T_A = T_A.copy()
-    T_A[np.isinf(T_A)] = np.nan
-    core.check_dtype(T_A)
-
-    if T_B is None:
-        T_B = T_A
-        ignore_trivial = True
-
-    T_B = np.asarray(T_B)
-    T_B = T_B.copy()
     if T_B.ndim != 1:  # pragma: no cover
         raise ValueError(
             f"T_B is {T_B.ndim}-dimensional and must be 1-dimensional. "
             "For multidimensional STUMP use `stumpy.mstump` or `stumpy.mstumped`"
         )
-    T_B[np.isinf(T_B)] = np.nan
+
+    core.check_dtype(T_A)
     core.check_dtype(T_B)
 
     core.check_window_size(m)
@@ -400,11 +414,6 @@ def stump(T_A, m, T_B=None, ignore_trivial=True):
     l = n - m + 1
     excl_zone = int(np.ceil(m / 4))  # See Definition 3 and Figure 3
 
-    M_T, Σ_T = core.compute_mean_std(T_A, m)
-    μ_Q, σ_Q = core.compute_mean_std(T_B, m)
-
-    T_A[np.isnan(T_A)] = 0
-
     out = np.empty((l, 4), dtype=object)
     profile = np.empty((l,), dtype="float64")
     indices = np.empty((l, 3), dtype="int64")
@@ -412,13 +421,10 @@ def stump(T_A, m, T_B=None, ignore_trivial=True):
     start = 0
     stop = l
 
-    profile[start], indices[start, :] = _get_first_stump_profile(
-        start, T_A, T_B, m, excl_zone, M_T, Σ_T, ignore_trivial
+    all_start_profiles, indices[start, :] = _get_first_stump_profile(
+        start, T_A, T_B, m, excl_zone, M_T, Σ_T, μ_Q, σ_Q, ignore_trivial
     )
-
-    T_B[
-        np.isnan(T_B)
-    ] = 0  # Remove all nan values from T_B only after first profile is calculated
+    profile[start] = all_start_profiles[0]
 
     QT, QT_first = _get_QT(start, T_A, T_B, m)
 
